/* s_seq.c (sequence, implemented as doubly linked list or as array)

   Ambros Marzetta
   
 * $Source: /home/schilte/marzetta/work/distr/zram/RCS/s_seq.c,v $
 * $Revision: 1.2 $
 * $Date: 1997/04/25 08:41:11 $
 * $Log: s_seq.c,v $
 * Revision 1.2  1997/04/25  08:41:11  marzetta
 * Added some type casts.
 *
 * Revision 1.1  1996/04/11  16:00:28  marzetta
 * Initial revision
 *
 */

#define _S_SEQ_C_
#include "s_seq.h"
#include <stdio.h>
#include <stdlib.h>		/* malloc */
#include "z_par.h"		/* assert */

/* Types ------------------------------------------------------------- */

struct s_seqT {
  long numElem;			/* number of elements in the sequence */
  s_seq_typeT type;
};

/* Doubly linked list implementation --------------------------------- */

typedef struct dl_refT {
  struct dl_refT *next, *prev;
  s_seq_elemP elem;
} *dl_refP;

typedef struct {
  struct s_seqT seq;	     /* common part for all sequence implementations */
  struct dl_refT anchor;
} *dl_seqP;

static dl_refP freeList;
static int freeListSize;

dl_seqP z_dl_New(void);
void z_dl_Free(dl_seqP l);
s_seq_elemP z_dl_GetObject(dl_seqP l, dl_refP where);
dl_refP z_dl_InsertFirst(dl_seqP l, s_seq_elemP x);
dl_refP z_dl_InsertLast(dl_seqP l, s_seq_elemP x);
dl_refP z_dl_InsertAfter(dl_seqP l, dl_refP r, s_seq_elemP x);
dl_refP z_dl_First(dl_seqP l);
dl_refP z_dl_Last(dl_seqP l);
void z_dl_ForAll(dl_seqP l, z_PtrProc whatToDo);
void z_dl_Delete(dl_seqP l, dl_refP ref);

dl_seqP z_dl_New(void)
{
  dl_seqP l;
  
  l = (dl_seqP) malloc(sizeof(*l));
  assert(l);
  l->anchor.next = & l->anchor;
  l->anchor.prev = & l->anchor;
  l->anchor.elem = NULL;
  return l;
} /* z_dl_New */

void z_dl_Free(dl_seqP l)
{
  free(l);
} /* z_dl_Free */

s_seq_elemP z_dl_GetObject(dl_seqP l, dl_refP where)
{
  return where->elem;
} /* z_dl_GetObject */

dl_refP z_dl_InsertFirst(dl_seqP l, s_seq_elemP x)
{
  dl_refP ref;

  if (freeList) {
    ref = freeList;
    freeList = freeList->next;
    freeListSize--;
  } else
    ref = (dl_refP) malloc(sizeof(*ref));
  assert(ref);
  ref->elem = x;
  ref->prev = & l->anchor;
  ref->next = l->anchor.next;
  l->anchor.next->prev = ref;
  l->anchor.next = ref;
  return ref;
} /* z_dl_InsertFirst */

dl_refP z_dl_InsertLast(dl_seqP l, s_seq_elemP x)
{
  dl_refP ref;

  if (freeList) {
    ref = freeList;
    freeList = freeList->next;
    freeListSize--;
  } else
    ref = (dl_refP) malloc(sizeof(*ref));
  assert(ref);
  ref->elem = x;
  ref->next = & l->anchor;
  ref->prev = l->anchor.prev;
  l->anchor.prev->next = ref;
  l->anchor.prev = ref;
  return ref;
} /* z_dl_InsertLast */

dl_refP z_dl_InsertAfter(dl_seqP l, dl_refP r, s_seq_elemP x)
{
/* Insert x into the sequence l after r. */
  assert(0);
  return NULL;
} /* z_dl_InsertAfter */

dl_refP z_dl_First(dl_seqP l)
{
  return l->anchor.next;
} /* z_dl_First */

dl_refP z_dl_Last(dl_seqP l)
{
  return l->anchor.prev;
} /* z_dl_Last */

void z_dl_ForAll(dl_seqP l, z_PtrProc whatToDo)
{
  dl_refP ref;

  for (ref = l->anchor.next; ref != &l->anchor; ref = ref->next)
    whatToDo(ref->elem);
} /* z_dl_ForAll */

void z_dl_Delete(dl_seqP l, dl_refP ref)
{
  ref->prev->next = ref->next;
  ref->next->prev = ref->prev;
  if (freeListSize < 30) {
    ref->next = freeList;
    freeList = ref;
    freeListSize ++;
  } else
    free(ref);
} /* z_dl_Delete */

/* Implementation table ---------------------------------------------- */

typedef s_seqP (*NewProc)(void);
static NewProc New[] = {(NewProc)z_dl_New};
     
typedef void (*FreeProc)(s_seqP s);
static FreeProc Free[] = {(FreeProc)z_dl_Free};

typedef s_seq_elemP (*GetProc)(s_seqP l, s_seq_refP where);
static GetProc Get[] = {(GetProc)z_dl_GetObject};

typedef s_seq_refP (*InsertProc)(s_seqP l, s_seq_elemP x);
static InsertProc InsertFirst[] = {(InsertProc)z_dl_InsertFirst},
                  InsertLast[] = {(InsertProc)z_dl_InsertLast};

typedef s_seq_refP (*InsertAfterProc)(s_seqP l, s_seq_refP r, s_seq_elemP x);
static InsertAfterProc InsertAfter[] = {(InsertAfterProc)z_dl_InsertAfter};

typedef s_seq_refP (*RefProc)(s_seqP l);
static RefProc First[] = {(RefProc)z_dl_First},
               Last[] = {(RefProc)z_dl_Last};

typedef void (*DeleteProc)(s_seqP l, s_seq_refP ref);
static DeleteProc Delete[] = {(DeleteProc)z_dl_Delete};

typedef void (*ForAllProc)(s_seqP l, z_PtrProc whatToDo);
static ForAllProc ForAll[] = {(ForAllProc)z_dl_ForAll};
     
/* Procedures (general)----------------------------------------------- */

s_seqP s_seq_New(s_seq_typeT type)
{
/* Allocate an empty list. Choose an implementation
   of type type. */
  s_seqP s;
  
  assert(type < s_seq_numtypes);
  s = New[type]();
  s->numElem = 0;
  s->type = type;
  return s;
} /* s_seq_New */


void s_seq_Free(s_seqP l)
{
/* Deallocate the sequence l, which must be empty. */
  assert(l->numElem == 0);
  Free[l->type](l);
} /* s_seq_Free */


s_seq_elemP s_seq_GetObject(s_seqP l, s_seq_refP where)
{
/* Retrieve the object referenced by where. */
  assert(l->type < s_seq_numtypes);
  return Get[l->type](l, where);
} /* s_seq_GetObject */


s_seq_refP s_seq_InsertFirst(s_seqP l, s_seq_elemP x)
{
/* Insert x into the sequence l (at the beginning). */
  l->numElem ++;
  return InsertFirst[l->type](l, x);
} /* s_seq_InsertFirst */


s_seq_refP s_seq_InsertLast(s_seqP l, s_seq_elemP x)
{
/* Insert x into the sequence l (at the end). */
  l->numElem ++;
  return InsertLast[l->type](l, x);
} /* s_seq_InsertLast */


s_seq_refP s_seq_InsertAfter(s_seqP l, s_seq_refP r, s_seq_elemP x)
{
/* Insert x into the sequence l after r. */
  l->numElem ++;
  return InsertAfter[l->type](l, r, x);
} /* s_seq_InsertAfter */


long s_seq_numElements(s_seqP l)
{
/* Return the number of elements in l. */
  return l->numElem;
} /* s_seq_numElements */


s_seq_refP s_seq_First(s_seqP l)
{
/* Return a reference to the first element of the sequence.
   The sequence must not be empty. */
  assert(l->numElem > 0);
  return First[l->type](l);
} /* s_seq_First */


s_seq_refP s_seq_Last(s_seqP l)
{
/* Return a reference to the last element of the sequence.
   The sequence must not be empty. */
  assert(l->numElem > 0);
  return Last[l->type](l);
} /* s_seq_Last */

void s_seq_ForAll(s_seqP l, z_PtrProc whatToDo)
{
  /* Call whatToDo for all elements of the sequence. WhatToDo must not
     access the sequence. */
  ForAll[l->type](l, whatToDo);
} /* s_seq_ForAll */

void s_seq_Delete(s_seqP l, s_seq_refP ref)
{
/* Delete the element referenced by ref. */
  l->numElem --;
  Delete[l->type](l, ref);
} /* s_seq_Delete */
