/* vc_base.c (Vertex cover problem)

   For more information see our WWW page at
   
         http://wwwjn.inf.ethz.ch/ambros/zram.html
	 
 * $Source: /home/schilte/marzetta/work/distr/zram/RCS/vc_base.c,v $
 * $Revision: 1.1 $
 * $Date: 1998/04/15 08:47:29 $
 * $Log: vc_base.c,v $
 * Revision 1.1  1998/04/15  08:47:29  marzetta
 * Initial revision
 *
 *
 */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "z_base.h"
#include "vc_base.h"

vc_GlobalP vc_NewFullNode(MINT numpt, MINT numedge)
{
/* Allocate a full vc_Node. Initialize len, tag, vc_numPt = vc_numAct = pt,
   vc_numMarked = 0, vc_numEdge = numedge; all odd points in left subgraph.
   */
  int len, i;
  vc_GlobalP np;

  assert(numpt < ADJ_SHIFT);
  assert(numedge + numpt < MAXMINT / ADJ_SHIFT);
  len = sizeof(vc_Global) + sizeof(MINT) * (5*numpt + numedge);
  np = (vc_GlobalP) z_NewNode(vc_GlobalTag, len);
  assert(np);
  if (np) {
    np->vc_numPt = numpt;
    np->vc_numAct = numpt;
    np->vc_numMarked = 0;
    np->vc_numEdge = numedge;
    np->vc_procsetnum = -1;
    np->vc_lbLeft = np->vc_lbRight = 0;
    for (i = 0; i < numpt; i++)
      vc_inleft(np)[i] = i % 2;
  } else
    puts("vc_NewFullNode: no memory");
  return np;
} /* vc_NewFullNode */

vc_NodeP vc_NewShortNode(MINT numpt, MINT numedge)
{
/* Allocate a vc_Node. Initialize len, tag, vc_numPt = vc_numAct = pt,
   vc_numMarked = 0, vc_numEdge = numedge;
   */
  int len;
  vc_NodeP np;

  assert(numpt < ADJ_SHIFT);
  assert(numedge + numpt < MAXMINT / ADJ_SHIFT);
  len = sizeof(vc_Node) + sizeof(degreeT) * numpt;
				/* This is one degreeT too much */
  np = (vc_NodeP) z_NewNode(vc_NodeTag, len);
  assert(np);
  if (np) {
    np->vc_numPt = numpt;
    np->vc_numAct = numpt;
    np->vc_numMarked = 0;
    np->vc_numEdge = numedge;
    np->vc_lb = 0;
  } else
    puts("vc_NewShortNode: no memory");
  return np;
} /* vc_NewShortNode */

void vc_FreeNode(vc_NodeP np)
{
  z_FreeNode((z_NodeP) np);
}

void vc_PrintNode(vc_GlobalP global, vc_NodeP node)
{
  /* Print node to stdout */
  int i;
  MINT *p, degree;

  assert(global && global->tag == vc_GlobalTag);
  printf("Node: len =%5ld\n",
	 (long) node->len);
  printf("Points:%7ld (active:%4ld, marked:%4ld)\nActive edges:%8ld\n",
	 (long) node->vc_numPt,
	 (long) node->vc_numAct,
	 (long) node->vc_numMarked,
	 (long) node->vc_numEdge);
  for (i=0;i<node->vc_numPt;i++) { /* one point's adjacency list */
    degree = vc_degree(node)[i];
    if (degree != MARK_DEGREE) {
      printf("%9d (%3ld):", i,
	     (long) degree);
      if (degree > 0) {
	p = vc_adjPt(global, i);
	while (*p != ADJ_LAST) {
	  if (vc_active(node, vc_pt(*p))) {
	    printf(" %1ld", (long)vc_pt(*p));
	    degree --;
	  }
	  p++;
	}
	if (degree)
	  printf(" ### bad degree ###");
      }
      printf("\n");
    } else
      printf("%9d *****\n", i);	/* marked */
  }
} /* vc_PrintNode */

void vc_Mark(vc_GlobalP global, vc_NodeP np, MINT pt, MINT **degree1list)
{
  /* Mark point pt and remove incident edges.
     Save points with degree 1 in degree1list if it is not NULL.
     */
  register MINT *adj;
  MINT degree;

  /*printf("    Marking %ld\n", (long) pt);*/
  assert(global && global->tag == vc_GlobalTag);
  assert(np && np->tag == vc_NodeTag);

  assert(pt < np->vc_numPt);
  degree = vc_degree(np)[pt];
  assert(degree > 0);
  np->vc_numEdge -= 2 * degree;	/* we remove two directed edges */
  vc_degree(np)[pt] = MARK_DEGREE;
  np->vc_numMarked ++;		/* This point becomes marked instead of*/
  np->vc_numAct --;		/* active */
  adj = vc_adjPt(global, pt);	/* head of pt's adjacency list */

#ifdef OLDMARK
  while (*adj != ADJ_LAST) {
    MINT other;
    other = vc_pt(*adj);	/* the adjacent point */
    if (vc_active(np, other)) { /* this edge is active */
      /*printf("Removing edge (%ld,%ld)\n", (long) pt, (long) other);*/
      degree--;
      if (--vc_degree(np)[other] == 0)
	np->vc_numAct --;	/* other point is no longer active */
      else
	if (vc_degree(np)[other] == 1 && degree1list) {
	  /*printf("      Degree 1: %ld\n", (long)other);*/
	  *((*degree1list)++) = other;
	}
    } /* other is active */
    adj++;
  } /* while */
  assert(degree == 0);
#else /* OLDMARK */
  while (*adj != ADJ_LAST) {
    register degreeT *deg;
    deg = vc_degree(np) + vc_pt(*adj); /* ptr to degree of adjacent point */
    if (*deg > 0) { /* this edge is active */
      if (--(*deg) <= 1)
	if (*deg == 0)
	  np->vc_numAct --;	/* other point is no longer active */
	else			/* new degree == 1 */
	  if (degree1list) *((*degree1list)++) = vc_pt(*adj);
    } /* this edge is active */
    adj++;
  } /* while */
#endif /* OLDMARK */

} /* vc_Mark */

MINT vc_MaxDegreePoint(vc_NodeP np)
{
  /* Return a point of maximum degree in np. */
  /* np must have an edge. */
  MINT maxdegree, pt, i;

  maxdegree = 0;
  pt = 0;			/* to avoid uninitialized variable warning */
  for (i=np->vc_numPt-1;i>=0;i--)
    if (vc_degree(np)[i] > maxdegree) {
      maxdegree = vc_degree(np)[i];
      pt = i;
    }
  assert(maxdegree > 0);
  return pt;

} /* vc_MaxDegreePoint */

boolean vc_IsSolution(vc_NodeP np)
{
  /* Is np a solution (leaf node of branch and bound tree)? */
  return np->vc_numAct == 0;
}

int vc_CompareProc(vc_NodeP x, vc_NodeP y)
{
  /* Compare the lower bounds of x and y */
  if (x->vc_lb > y->vc_lb) return 1;
  else if (x->vc_lb == y ->vc_lb) return 0;
  else return -1;
}

char *vc_ShortNodeString(vc_NodeP np)
{
  /* Return a static string describing np */
  static char nodeString[80];

  if (np) {
    sprintf(nodeString, "e=%ld, act=%ld, marked=%ld, lb=%ld",
	    (long) np->vc_numEdge,
	    (long) np->vc_numAct,
	    (long) np->vc_numMarked,
	    (long) np->vc_lb);
    return nodeString;
  } else
    return "NULL Node";
}

void MarkHeadTail(vc_GlobalP global, vc_NodeP np,
			 MINT **head, MINT **tail)
{
  /* Mark partners of points in list. The points in the list have degree
     at most one.
     */
  MINT other;
  
  while (*head != *tail) {
    if (vc_active(np, **tail)) {
      other = SomeAdjacent(global, np, **tail);
      vc_Mark(global, np, other, head);
    }
    (*tail) ++;
  }
} /* MarkHeadTail */

MINT SomeAdjacent(vc_GlobalP global, vc_NodeP np, MINT pt)
{
  /* Return some point adjacent to pt. */
  register MINT *adj;

  assert(vc_active(np, pt));
  adj = vc_adjPt(global, pt);
  for (;;) {
    if vc_active(np, vc_pt(*adj)) return vc_pt(*adj);
    adj ++;
  }
} /* SomeAdjacent */

void UnmarkPoints(vc_GlobalP global, vc_NodeP np)
{
  /* np has no active points. Unmark points if possible. */
  MINT pt, *adj;
  boolean unmarkedNeighbor;
  
  for (pt = np->vc_numPt; pt >= 0; pt --)
    if (vc_degree(np)[pt] == MARK_DEGREE) {
      adj = vc_adjPt(global, pt);
      unmarkedNeighbor = 0;
      while (! unmarkedNeighbor && *adj != ADJ_LAST) {
	unmarkedNeighbor = vc_degree(np)[vc_pt(*adj)] == 0;
	adj++;
      } /* while */
      if (! unmarkedNeighbor) {
	vc_degree(np)[pt] = 0;	/* inactive, not marked */
	np->vc_numMarked --;
	/* printf("Unmark %ld\n", (long) pt); */
      } /* pt has only marked neighbors */
    } /* pt is marked*/
} /* UnmarkPoints */

void Restrict(vc_GlobalP global, vc_NodeP np, int left)
{
  /* Restrict np to the left / right subgraph. Degree of points in the
     other subgraph is set to 0. Set lb to 0.
     */
  int pt, edges;
  MINT *adj;
  degreeT *degree;

  /*printf("Restrict: %ld ", (long) np->vc_numAct);*/
  np->vc_numAct = np->vc_numMarked = 0;
  edges = 0;
  for (pt = 0; pt < np->vc_numPt; pt ++)
    if (vc_inleft(global)[pt] == left) {
      if (vc_active(np, pt)) {
	degree = &(vc_degree(np)[pt]);
	*degree = 0;
	/* count edges */
	for (adj = vc_adjPt(global, pt); *adj != ADJ_LAST; adj++)
	  if (vc_inleft(global)[vc_pt(*adj)] == left &&
	      vc_active(np, vc_pt(*adj))) {
	    /*printf("    Edge %d,%d\n", pt, vc_pt(*adj));*/
	    (*degree) ++;
	    edges ++;
	  } /* for, if */
	if (*degree) np->vc_numAct ++;
      } else if (vc_degree(np)[pt] == MARK_DEGREE)
	np->vc_numMarked ++;
    } else			/* pt is in other subgraph */
      vc_degree(np)[pt] = 0;	/* inactive */
  np->vc_numEdge = edges;
  np->vc_lb = 0;
  /*printf("%ld\n", (long) np->vc_numAct);*/
} /* Restrict */

void ForAllTriangles(vc_GlobalP global, vc_NodeP np,
		     TriangleProcP what, void *ref)
{
  /* Try a partition of np into triangles.
     Call what(ref, a, b, c) for all triangles found.
     */
  MINT i, j, k, a, b, p, q;
  MINT *adji, *adjb, *adjp;
  boolean found;

  for (i = 0; i<np->vc_numPt; i ++)
    if (vc_degree(np)[i] >= 2) {
      found = 0;
      adji = vc_adjPt(global, i);
      while (!found && *adji != ADJ_LAST) {
	j = vc_pt(*adji);
	assert(0 <= j && j < np->vc_numPt);
	if (j > i && vc_degree(np)[j] >= 2) { /* process every edge once */
	  if (vc_degree(np)[i] < vc_degree(np)[j])
	    {a = i; b = j;}
	  else
	    {a = j; b = i;}
	  /* degree(a) <= degree(b) */
	  adjb = vc_adjPt(global, b);
	  while (!found && *adjb != ADJ_LAST) {
	    k = vc_pt(*adjb);
	    if (vc_degree(np)[k] >= 2) {
	      /* There are edges (a,b) and (b,k) */
	      if (k > j) {	/* process every triangle once */
		if (vc_degree(np)[a] < vc_degree(np)[k])
		  {p = a; q = k;}
		else
		  {p = k; q = a;}
		/* degree(p) <= degree(q) */
		adjp = vc_adjPt(global, p);
		while (!found && *adjp != ADJ_LAST) {
		  if (vc_pt(*adjp) == q) {
		    /*printf("Triangle: %ld %ld %ld\n",
		      (long)i, (long)j, (long)k);*/
		    what(ref, i, j, k);
		    found = 1;
		  }
		  adjp ++;
		} /* while adjp */
	      } /* k > j */
	    } /* degree(k) >= 2 */
	    adjb ++;
	  } /* while *adjb */
	} /* if j */
	adji ++;
      } /* while *adji */
    } /* for i, if degree(i) */
} /* ForAllTriangles */
