/* vc_lb.c

   Vertex cover lower bounds

   Ambros Marzetta

   94-12-13
   For more information see our WWW page at
   
         http://wwwjn.inf.ethz.ch/ambros/zram.html
	 
 * $Source: /home/schilte/marzetta/work/distr/zram/RCS/vc_lb.c,v $
 * $Revision: 1.1 $
 * $Date: 1998/04/15 08:47:29 $
 * $Log: vc_lb.c,v $
 * Revision 1.1  1998/04/15  08:47:29  marzetta
 * Initial revision
 *
 *
 */

#include <stdio.h>
#include "z_bb.h"
#include "vc_base.h"
#include "vc_lb.h"

#define ALLTRIANGLES		/* saves 9% nodes, 5% time in vc9 */
#define HISTO0
#define TRACELB0
#define NOT_MAXIMAL0		/* only BetterQuickMatch, not Edmonds */
#define BREADTH_FIRST		/* algorithm of Edmonds, not depth first*/
#define NO_CIRCLES		/* do not search circles after Edmonds */

static MINT *degree1list;	/* used by LowerBound234 */
static long degree1listsize;

/* Lower Bound 1: useless lower bound ------------------------------------ */

MINT vc_LowerBound1(vc_GlobalP global, vc_NodeP np)
{
/* The simplest of all lower bounds (edges / maxdegree), not useful at all */
  MINT maxdegree;

  if (np->vc_numEdge == 0)
    return np->vc_numMarked;	/* This is a solution node */
  
  maxdegree = vc_degree(np)[vc_MaxDegreePoint(np)];

  return np->vc_numMarked + (np->vc_numEdge/2 + maxdegree-1) / maxdegree;
			/* numEdge/2 because it is an undirected graph */
}

/* Lower Bound 7, 8: maximum matching ------------------------------------ */

#define NO_ODD_CYCLE (-1)	/* must be < 0 */

static char *state;		/* 0..7: 1=odd, 2=even dist from root */
				/* 4=on current path from root to cur */
static long statesize;
static int *neighbour;
static long neighboursize;

/* values for neighbour: */
#define INACTIVE (-3)
#define IN_CIRCLE (-2)
#define UNMATCHED (-1)
/* order is significant: neighbour[i] >= UNMATCHED is a possible question */

static boolean FindAltPathDF(vc_GlobalP global, vc_NodeP np, int cur)
{
  /* Depth first.
     There is an alternating path from root to cur.
     Cur is at an even distance from root.
     Last edge (___ -> cur) is marked (if the path is not empty).
     Next edge (cur -> next) should be unmarked.
     Return true if path could be completed to an unmarked point.
     */
  MINT *adj;
  int next;

  assert(neighbour[cur] >= UNMATCHED); /* == UNMATCHED only if cur == root */
  if ((state[cur] & (2+4)) == 0) {	/* not yet visited in this state */
    state[cur] |= 2+4;
    adj = vc_adjPt(global, cur);
    while (*adj != ADJ_LAST) {
      next = vc_pt(*adj);
      if (neighbour[next] >= UNMATCHED &&
	  ((state[next] & (1+4)) == 0)) {
	state[next] |= 1+4;
	if (neighbour[next] == UNMATCHED ||
	    FindAltPathDF(global, np, neighbour[next])) {
	  neighbour[next] = cur;
	  neighbour[cur] = next;
	  return 1;
	}
	state[next] &= 3;
      } /* if */
      adj ++;
    } /* while *adj */
    state[cur] &= 3;
  } /* state[cur] & 2 == 0 */
  return 0;			/* did not find an alternating path */
} /* FindAltPathDF */

/* Algorithm of Edmonds
   There are real points (number < np->vc_numPt) and non-real points (blossom).
   If a point p is part of a blossom, its number is blossom[p], otherwise
   blossom[p] == -1.
   Every point has a state (bstate[p]).
       inactive	-> not part of the graph which is input to the algorithm.
       undef	-> not yet visited
       odd	-> at odd distance from some root and not part of a blossom
       even	-> at even distance from some root (maybe part of a blossom)
       blossom	-> not a real point
   The queue (queue[head]..queue[tail]) contains only real points p
   with bstate[p] == even.
   For every even real points and every blossom there is a path back to a root:
       back[p]		-> real point or blossom with (base1[back[p]] != -1)
       doubleback[p]	-> real point (even)
       dist[p]		-> dist[doubleback[blossom[p]]] < dist[p]
   The edge (p, back[p]) is always marked. The edge (back[p], doubleback[p])
   is never marked. If back[p] is a blossom, the path is (p, base[back[p]],
   ..., base1[back[p]], doubleback[p], ...).
   For every blossom there is a real point base[b]. For every blossom b with
   (b == back[p] for some p) (this means b is part of another blossom and
   not at its base) there is a real point base1[b].
   */

#define toblossom(v) while (blossom[v] != -1) v = blossom[v]
#define MAXPT 1000

static int head, tail;		/* queue for breadth first search */
static int queue[MAXPT],		/* contains real points */
      blossom[MAXPT],
      back[MAXPT], doubleback[MAXPT], dist[MAXPT], /* even points */
      base[MAXPT], base1[MAXPT];
static enum {inactive, undef, odd, even, bluet} bstate[MAXPT]; /* all points */

void ProcessHalfBlossom(int v0, int lastv0, int baseblossom, int newblossom)
{
  /* Process path backwards from v0 (even) to baseblossom (even).
     v0 is an even real point which may be part of a blossom.
     v1 is an odd real point.
     Update blossom[] for all points on the path.
     Update back[], doubleback[], dist[] for odd points.
     Update base1[] for every blossom.
     Insert odd points into queue and make them even.
     */
  int v0blossom, v1;

  assert(bstate[v0] == even);
  v0blossom = v0;
  toblossom(v0blossom);
  while (v0blossom != baseblossom) {
    /*printf("    return + %d %d\n", v0, v0blossom);*/
    assert(v0blossom != newblossom);
    blossom[v0blossom] = newblossom;

    v1 = back[v0blossom];
    assert(bstate[v1] == odd);
    assert(blossom[v1] == -1);
    /*printf("    return - %d\n", v1);*/
    blossom[v1] = newblossom;

    if (v0blossom == v0) { /* real point */
      back[v1] = v0;
    } else {		/* v0 is (in) blossom */
      assert(base1[v0blossom] == -1);
      base1[v0blossom] = v0;
      back[v1] = v0blossom;
    }
    doubleback[v1] = lastv0;
    dist[v1] = dist[newblossom];
    queue[tail++] = v1;
    bstate[v1] = even;
    lastv0 = v1;
    v0 = doubleback[v0blossom];
    assert(bstate[v0] == even);
    v0blossom = v0;
    toblossom(v0blossom);
  } /* while v0blossom != baseblossom */
} /* ProcessHalfBlossom */

void InvertBlossom(int bl, int numPt)
{
  /* Invert the path between base[bl] and base1[bl].
     This procedure is rarely called.
     */
  int v0;

  /*printf("[");*/
  v0 = base1[bl];
  while (v0 != base[bl]) {
    /*printf("%4d%4d", back[v0], doubleback[v0]);*/
    if (back[v0] >= numPt) { /* Blossom */
      InvertBlossom(back[v0], numPt);
      neighbour[base1[back[v0]]] = doubleback[v0];
      neighbour[doubleback[v0]] = base1[back[v0]];
    } else {
      neighbour[back[v0]] = doubleback[v0];
      neighbour[doubleback[v0]] = back[v0];
    }
    v0 = doubleback[v0];
  }
  /*printf("]");*/
} /* InvertBlossom */

static boolean FindAltPathBF(vc_GlobalP global, vc_NodeP np, int root)
{
  /* see Ottmann / Widmayer, p. 612 */
  int i, count;
  
  head = tail = 0;
  for (i = np->vc_numPt - 1; i >= 0; i--) {
    blossom[i] = -1;
    if (neighbour[i] > UNMATCHED)
      bstate[i] = undef;
    else if (neighbour[i] == UNMATCHED) {
      bstate[i] = even;
      queue[tail++] = i;	/* insert into queue */
      dist[i] = 0;		/* necessary only for even nodes */
      back[i] = -1;		/* real points */
      doubleback[i] = -1;	/* real points */
    } else
      bstate[i] = inactive;
  } /* for i */
  count = np->vc_numPt;		/* number of next blossom */

  while (head != tail) { /* while queue not empty */
    MINT *adj;
    int cur, next, curblossom, nextblossom, v0, v1;

    cur = queue[head++];	/* cur is at even distance from some root */
    assert(bstate[cur] == even);
    assert(cur < np->vc_numPt);	/* cur is a real point */
    curblossom = cur;
    toblossom(curblossom);
    adj = vc_adjPt(global, cur);
    while (*adj != ADJ_LAST) {
      next = vc_pt(*adj);
      nextblossom = next;
      toblossom(nextblossom);
      if (nextblossom != curblossom && /* next must not be in same blossom */
	  next != neighbour[cur]) { /* edge must be unmarked */
	if (bstate[next] == undef) { /* case 2 */
	  /*printf("  Add %3d %3d %3d\n", cur, next, neighbour[next]);*/
	  bstate[next] = odd;
	  assert(neighbour[next] > UNMATCHED);
	  assert(bstate[neighbour[next]] == undef);
	  bstate[neighbour[next]] = even;
	  back[neighbour[next]] = next;
	  doubleback[neighbour[next]] = cur;
	  dist[neighbour[next]] = dist[cur] + 1;
	  queue[tail++] = neighbour[next];
	} else if (bstate[next] == even) {	/* cases 3 and 4 */
	  /* find common ancestor if it exists */
	  v0 = cur;
	  toblossom(v0);
	  v1 = next;
	  toblossom(v1);
	  while (v0 != v1 && (dist[v0] || dist[v1])) {
	    if (dist[v0] > dist[v1]) {
	      v0 = doubleback[v0];
	      toblossom(v0);
	      /*printf("%d ", v0);*/
	    } else {
	      v1 = doubleback[v1];
	      toblossom(v1);
	      /*printf("%d ", v1);*/
   	    }
	  }
	  if (v0 == v1) {      /* case 3: there is a common ancestor: blossom */
	    int mybase, baseblossom;
	    baseblossom = v0;
	    mybase = v0;	/* may be another blossom */
	    if (mybase >= np->vc_numPt) mybase = base[mybase]; /* real point */
	    /*printf("  Blossom %d at %3d %3d (%3d %3d), base = %d (%d)\n",
		   count, cur, next, curblossom, nextblossom,
		   mybase, baseblossom);*/
	    assert(bstate[mybase] == even);

	    /* Create a new blossom with base mybase and number count */
	    
	    base[count] = mybase;
	    base1[count] = -1;
	    back[count] = back[mybase];
	    doubleback[count] = doubleback[mybase];
	    dist[count] = dist[mybase];
	    blossom[count] = -1;
	    bstate[count] = bluet; /* necessary? */

	    ProcessHalfBlossom(cur, next, baseblossom, count); /* cur side */
	    ProcessHalfBlossom(next, cur, baseblossom, count); /* next side */
	    
	    blossom[baseblossom] = count;
	    curblossom = count;
	    count ++;
	  } else {		/* case 4: found a path */
	    /*printf("  Path %3d %3d (roots %3d %3d)", cur, next, v0, v1);*/
	    neighbour[cur] = next;
	    neighbour[next] = cur;

	    v0 = cur;
	    while (back[v0] >= 0) {
	      /*printf("%4d%4d", back[v0], doubleback[v0]);*/
	      if (back[v0] >= np->vc_numPt) { /* Blossom */
		InvertBlossom(back[v0], np->vc_numPt);
		neighbour[base1[back[v0]]] = doubleback[v0];
		neighbour[doubleback[v0]] = base1[back[v0]];
	      } else {
		neighbour[back[v0]] = doubleback[v0];
		neighbour[doubleback[v0]] = back[v0];
	      }
	      v0 = doubleback[v0];
	    } /* while */
	    
	    v0 = next;
	    while (back[v0] >= 0) {
	      /*printf("%4d%4d", back[v0], doubleback[v0]);*/
	      if (back[v0] >= np->vc_numPt) { /* Blossom */
		InvertBlossom(back[v0], np->vc_numPt);
		neighbour[base1[back[v0]]] = doubleback[v0];
		neighbour[doubleback[v0]] = base1[back[v0]];
	      } else {
		neighbour[back[v0]] = doubleback[v0];
		neighbour[doubleback[v0]] = back[v0];
	      }
	      v0 = doubleback[v0];
	    } /* while */

	    /*printf("\n");*/
	    return 1;
	  } /* case 4 */
	} /* cases 3 and 4 */
      } /* nextblossom != curblossom */

      adj ++;
    } /* while *adj */
  } /* while queue not empty */
  /* queue is empty: there is no alternating path */
  /*printf("--No path\n");*/
  return 0;
} /* FindAltPathBF */

static int degree0[MAXPT];	/* for BetterQuickMatch and Match */
#define degree (degree0 + 1)

static void Match(vc_GlobalP global, int *neighbour, int *delta,
	     int i, int other, int **tail, MINT *adji)
{
  /* This procedure is called only from BetterQuickMatch.
     Match i and other.
     Update neighbour, increase delta, update degree.
     Put new degree 1 points into *tail.
     adji == NULL -> i has no other neighbours.
     adji != NULL -> this is the first edge from i to a point with degree != 0
     */
  register MINT *adj2;
  
  neighbour[other] = i;
  neighbour[i] = other;
  (*delta) ++;
  degree[i] = 0;
  degree[other] = 0;
  if (adji) {
    /*adj2 = vc_adjPt(global, i);*/
    adj2 = adji + 1;
    while (*adj2 != ADJ_LAST) {
      register int *deg;
      
      deg = degree + vc_pt(*adj2);
      if (*deg > 0) {
	(*deg) --;
	if (*deg == 1) *((*tail)++) = vc_pt(*adj2);
      }
      adj2 ++;
    } /* while */
  }
  adj2 = vc_adjPt(global, other);
  while (*adj2 != ADJ_LAST) {
    register int *deg;

    deg = degree + vc_pt(*adj2);
    if (*deg > 0) {
      (*deg) --;
      if (*deg == 1) *((*tail)++) = vc_pt(*adj2);
    }
    adj2 ++;
  } /* while */
} /* Match */

static void BetterQuickMatch(vc_GlobalP global, vc_NodeP np,
		       int *neighbour, int *delta)
{
  /* Find quickly an initial matching. This procedure could be left
     away. It only makes the computation faster.
     Time = O(n+e) */
  int i;
  int degree1list[MAXPT], *head, *tail;

  degree[-1] = -1;
  for (i = np->vc_numPt - 1; i >= 0; i--)
    degree[i] = (neighbour[i] == UNMATCHED) ? vc_degree(np)[i] : 0;

  head = tail = degree1list;
  
  for (i = np->vc_numPt - 1; i >= 0; i--)
    if (neighbour[i] == UNMATCHED) { /* try to match point i */
      register MINT *adj;
      int other;
      
      adj = vc_adjPt(global, i);
      /*while (*adj != ADJ_LAST && degree[vc_pt(*adj)] == 0) adj ++;*/
      while (degree[vc_pt(*adj)] == 0) adj ++;
      if (*adj != ADJ_LAST) {	/* found a matching between i and other */
	other = vc_pt(*adj);
	Match(global, neighbour, delta, i, other, &tail, adj);
	while (head != tail) {
	  if (degree[*head] > 0) {
	    register MINT *adj2;
	    assert(degree[*head] == 1);
	    adj2 = vc_adjPt(global, *head);
	    while (degree[vc_pt(*adj2)] == 0) adj2 ++;
	    Match(global, neighbour, delta, *head, vc_pt(*adj2), &tail, 0);
	  }
	  head ++;
	} /* while */
      } /* if */
    } /* for */
} /* BetterQuickMatch */

static MINT vc_LowerBound7(vc_GlobalP global, vc_NodeP np)
{
  /* maximum matching */
  int i, delta;

  if (np->vc_numEdge == 0)
    return np->vc_numMarked;	/* This is a solution node */

  z_ReSizePointer((void *)&state, &statesize, np->vc_numPt);
  z_ReSizePointer((void *)&neighbour, &neighboursize,
		  np->vc_numPt * sizeof(int));

  /* start with the empty matching */
  for (i = np->vc_numPt - 1; i >= 0; i--)
    neighbour[i] = vc_active(np, i) ? UNMATCHED : INACTIVE;
  delta = 0;			/* (# entries in neighbour >= 0) / 2 */

  BetterQuickMatch(global, np, neighbour, &delta);
  
#ifndef NOT_MAXIMAL

  /*printf("delta = %d\n", delta);*/
#ifdef BREADTH_FIRST
  /* find an alternating path starting at root */
  while ((delta + 1) * 2 <= np->vc_numAct && FindAltPathBF(global, np, 0)) {
    delta ++;
  }
#else /* BREADTH_FIRST */
  for (root = np->vc_numPt - 1;
       (delta + 1) * 2 <= np->vc_numAct && root >= 0;
       root --)
    if (neighbour[root] == UNMATCHED) {
      /* find an alternating path starting at root */
      /*printf("root = %d\n", root);*/
      for (i = np->vc_numPt-1; i >= 0; i--) state[i] = 0;
      if (FindAltPathDF(global, np, root)) {
	delta ++;
      }
    } /* for root */
#endif /* BREADTH_FIRST */
  /*printf("delta = %d\n", delta);*/
  
#endif /* NOT_MAXIMAL */

  if (delta * 2 > np->vc_numAct)
    printf("ddd delta=%d, numAct=%ld, numMarked=%ld\n",
	   delta, (long)np->vc_numAct, (long)np->vc_numMarked);
  assert(delta * 2 <= np->vc_numAct);
  return np->vc_numMarked + delta;
} /* vc_LowerBound7 */

static int altCycles;

static int FindAltCycles(vc_GlobalP global, vc_NodeP np, int cur, int root)
{
  /* Path from root to cur-1 has no cycles. Test whether cur is on the path.
     Last edge is marked (if the path is not empty).
     Edge from cur to next should be unmarked.
     Return NO_ODD_CYCLE if no odd cycle is found.
     Return point number of start of cycle otherwise.
     !!! Finds only long circles !!!
     state: 0=unknown, 1=lastmarked, 2=lastunmarked, 3=othercycle
     */
  int next, start;
  MINT *adj;
  
  printf("root = %d\n", root);
  assert(state[cur] == 0 || state[cur] == 3);
  assert(neighbour[cur] != IN_CIRCLE);
  if (state[cur] == 0) {
    state[cur] = 1;
    adj = vc_adjPt(global, cur);
    while (*adj != ADJ_LAST) {
      next = vc_pt(*adj);
      assert(next != cur);	/* no length 1 cycles */
      if (neighbour[next] != IN_CIRCLE &&
	  next != neighbour[cur] && /* next edge should be unmarked */
	  vc_active(np, next)) {
	if (next == root) { /* found odd cycle */
	  state[cur] = 3;
	  state[next] = 3;
	  if (neighbour[root] >= 0)
	    neighbour[neighbour[root]] = -1;
	  else
	    altCycles ++;
	  neighbour[cur] = IN_CIRCLE;
	  neighbour[next] = IN_CIRCLE;
	  printf("***%3d%3d", next, cur);
	  return next;
	} else if (neighbour[next] >= 0 &&
		   state[next] == 0) { /* not yet visited */
	  state[next] = 2;
	  if (neighbour[next] != root)
	    start = FindAltCycles(global, np, neighbour[next], root);
	  else
	    start = NO_ODD_CYCLE; /* even cycle */
	  state[next] = 3;
	  if (start != NO_ODD_CYCLE) {
	    printf("%3d%3d", next, cur);
	    neighbour[cur] = IN_CIRCLE;
	    neighbour[next] = IN_CIRCLE;
	    state[cur] = 3;
	    if (start != cur)
	      return start;
	    else
	      return NO_ODD_CYCLE;
	  }
	}
      }
      adj ++;
    } /* while */
    state[cur] = 3;
  }
  return NO_ODD_CYCLE;
} /* FindAltCycles */

/*int back[MAXPT];
int queue[MAXPT];*/
boolean possiblecycleroot[MAXPT];

#ifndef NO_CIRCLES

static void FindAltCyclesBF(vc_GlobalP global, vc_NodeP np, int root)
{
  int head, tail, cur, i, next;
  MINT *adj;

  for (i = np->vc_numPt-1; i >= 0; i--) state[i] = 0;
  assert(vc_active(np, root));
  assert(neighbour[root] != IN_CIRCLE);
  for (i = global->vc_numPt - 1; i >= 0; i --) back[i] = -1;
  head = 0;
  tail = 0;
  queue[tail++] = root;
  state[root] = 2;
  while (head < tail) {
    cur = queue[head++];
    /* edge from cur to next must be unmarked */
    adj = vc_adjPt(global, cur);
    while (*adj != ADJ_LAST) {
      next = vc_pt(*adj);
      if ((state[next] & 1) == 0 && /* no even circle */
		 next != neighbour[cur] && /* edge unmarked */
		 neighbour[next] > UNMATCHED) {
	if (state[next] == 2) { /* odd circle */
	  /* next may be root */
#ifdef TRACELB
	  printf("Odd circle:");
#endif /* TRACELB */
	  i = next;
	  while(i != root) {
	    state[i] = 8;
	    i = back[i];
	  }
	  state[i] = 8;		/* state[root] */
	  i = cur;
	  while (state[i] != 8) {
#ifdef TRACELB
	    printf("%3d", i);
#endif /* TRACELB */
	    neighbour[i] = IN_CIRCLE;
	    i = back[i];
	  }
	  /* i == base */
	  assert(possiblecycleroot[i]);
	  assert(i <= root);
	  if (i == root && neighbour[root] == -1)
	    altCycles ++;
	  neighbour[neighbour[i]] = -1;
	  neighbour[i] = IN_CIRCLE;
	  i = next;
	  while (neighbour[i] != IN_CIRCLE) {
#ifdef TRACELB
	    printf("%3d", i);
#endif /* TRACELB */
	    neighbour[i] = IN_CIRCLE;
	    i = back[i];
	  }
#ifdef TRACELB
	    printf("%3d\n", i);
#endif /* TRACELB */
	  return;      /* found a cycle; there may be more */
	} /* odd circle */
	state[next] |= 1;
	back[next] = cur;
	assert(vc_active(np, neighbour[next]));
	queue[tail++] = neighbour[next];
	state[neighbour[next]] |= 2;
	back[neighbour[next]] = next;
      }
      adj ++;
    }
  } /* while head < tail */
  /* did not find any cycles */
  for (i = np->vc_numPt-1; i >= 0; i--)
    if (state[i] == 2)
      possiblecycleroot[i] = 0;
} /* FindAltCyclesBF */

#endif /* NO_CIRCLES */

static MINT vc_LowerBound8(vc_GlobalP global, vc_NodeP np)
{
  /* maximum matching and some odd circles*/
  int delta;
  
  if (np->vc_numEdge == 0)
    return np->vc_numMarked;	/* This is a solution node */

  delta = vc_LowerBound7(global, np); /* includes vc_numMarked */
  /* neighbour[] contains the matching. */

  altCycles = 0;
#ifdef AA
  for (i = np->vc_numPt-1; i >= 0; i--)
    printf("%3d-%3d\n", i, neighbour[i]);
#endif

#ifndef NO_CIRCLES
  for (i = np->vc_numPt-1; i >= 0; i--)
    possiblecycleroot[i] = 1;

  for (root = np->vc_numPt - 1; root >= 0; root --)
    while (neighbour[root] >= UNMATCHED && possiblecycleroot[root]) {
      /* find an alternating cycle starting/ending at root */
      /*DF: !!!!! for (i = np->vc_numPt-1; i >= 0; i--) state[i] = 0;*/
     FindAltCyclesBF(global, np, root);
    } /* for root */

#ifdef AA
  for (i = np->vc_numPt-1; i >= 0; i--)
    printf("%3d-%3d\n", i, neighbour[i]);
#endif

  /* matching again */
  for (root = np->vc_numPt - 1; root >= 0; root --)
    if (neighbour[root] == UNMATCHED) {
      /* find an alternating path starting at root */
      /*printf("root = %d\n", root);*/
      for (i = np->vc_numPt-1; i >= 0; i--) state[i] = 0;
      if (FindAltPathDF(global, np, root)) {
	delta ++;
      }
    } /* for root */
#endif /* NO_CIRCLES */

#ifdef TRACELB
  for (i = np->vc_numPt-1; i >= 0; i--)
    printf("%3d-%3d\n", i, neighbour[i]);
#endif /* TRACELB */

  return delta + altCycles;
} /* vc_LowerBound8 */

static MINT LowerBoundRest(vc_GlobalP global, vc_NodeP np,
			   MINT *head, MINT *tail)
{
  int lb;
  lb = vc_LowerBound8(global, np);

  vc_FreeNode(np);
  
  return lb;
} /* LowerBoundRest */

MINT vc_LowerBound2(vc_GlobalP global, vc_NodeP np)
{
  /* Maximal matching */
  MINT *head, *tail;
  
  if (np->vc_numEdge == 0)
    return np->vc_numMarked;	/* This is a solution node */

  z_ReSizePointer((void *)&degree1list,
		&degree1listsize, np->vc_numPt * sizeof(MINT));
  head = tail = degree1list;
  np = (vc_NodeP) z_NewCopyNode((z_NodeP) np);
  return LowerBoundRest(global, np, head, tail);
}

/* Lower bound 3: find some triangles ------------------------------------ */

MINT vc_LowerBound3(vc_GlobalP global, vc_NodeP np)
{
  /* Triangles */
  MINT i, j, k, a, b, p, q;
  MINT *adji, *adjb, *adjp;
  int numTriangle;
  MINT *head, *tail;
  boolean found;

  if (np->vc_numEdge == 0)
    return np->vc_numMarked;	/* This is a solution node */

  z_ReSizePointer((void *)&degree1list,
		&degree1listsize, np->vc_numPt * sizeof(MINT));
  head = tail = degree1list;
  np = (vc_NodeP) z_NewCopyNode((z_NodeP) np);

  numTriangle = 0;
  for (i = 0; i<np->vc_numPt; i ++)
    if (vc_degree(np)[i] >= 2) {
      found = 0;
      adji = vc_adjPt(global, i);
      while (!found && *adji != ADJ_LAST) {
	j = vc_pt(*adji);
	assert(0 <= j && j < np->vc_numPt);
	if (j > i && vc_degree(np)[j] >= 2) { /* process every edge once */
	  if (vc_degree(np)[i] < vc_degree(np)[j])
	    {a = i; b = j;}
	  else
	    {a = j; b = i;}
	  /* degree(a) <= degree(b) */
	  adjb = vc_adjPt(global, b);
	  while (!found && *adjb != ADJ_LAST) {
	    k = vc_pt(*adjb);
	    if (vc_degree(np)[k] >= 2) {
	      /* There are edges (a,b) and (b,k) */
	      if (k > j) {	/* process every triangle once */
		if (vc_degree(np)[a] < vc_degree(np)[k])
		  {p = a; q = k;}
		else
		  {p = k; q = a;}
		/* degree(p) <= degree(q) */
		adjp = vc_adjPt(global, p);
		while (!found && *adjp != ADJ_LAST) {
		  if (vc_pt(*adjp) == q) {
#ifdef TRACELB
		    printf("Triangle: %ld %ld %ld\n",
		      (long)i, (long)j, (long)k);
#endif
		    vc_Mark(global, np, i, &head);
		    vc_Mark(global, np, j, &head);
		    if (vc_degree(np)[k] > 0) {
		      vc_Mark(global, np, k, &head);
		      numTriangle ++;
		    }
		    MarkHeadTail(global, np, &head, &tail);
		    found = 1;
		  }
		  adjp ++;
		} /* while adjp */
	      } /* k > j */
	    } /* degree(k) >= 2 */
	    adjb ++;
	  } /* while *adjb */
	} /* if j */
	adji ++;
      } /* while *adji */
    } /* for i, if degree(i) */
  
  return LowerBoundRest(global, np, head, tail) - numTriangle;
} /* vc_LowerBound3 */

/* Lower bound 4: find some cliques --------------------------------------- */

MINT vc_LowerBound4(vc_GlobalP global, vc_NodeP np)
{
  /* Find cliques. This corresponds to finding a coloring of the
     complement graph. Lower bound = n - chi(complement).
     */
  MINT *head, *tail;
  int i, j, k, nextPt, delta;
  static MINT **adjlist;
  static int *member;
  static long adjlistsize, membersize;

  if (np->vc_numEdge == 0)
    return np->vc_numMarked;	/* This is a solution node */

  assert(global);
  z_ReSizePointer((void *)&adjlist, &adjlistsize,
		global->vc_numPt * sizeof(MINT *));
  z_ReSizePointer((void *)&member, &membersize,
		global->vc_numPt * sizeof(int));
  z_ReSizePointer((void *)&degree1list,
		&degree1listsize, np->vc_numPt * sizeof(MINT));
  head = tail = degree1list;
  np = (vc_NodeP) z_NewCopyNode((z_NodeP) np);
  delta = 0;

  for (i=0; i < global->vc_numPt; i++)
    if (vc_degree(np)[i] >= 2) {
      /* find clique starting with i and mark it */
      /* executed 25 times */
      /*printf("Start with %d\n", i);*/
      member[0] = i;
      adjlist[0] = vc_adjPt(global, i);
      k = 1;			/* number of clique members so far */
      /*printf("%4ld", (long) vc_triangle(global)[i]);*/
      if (vc_triangle(global)[i] == -1)
	nextPt = i + 1;
      else
	nextPt = vc_triangle(global)[i];
      assert(nextPt > i);
      
      for (;;) { /* Try some second point */
	/* executed 27 times */
	
	for (;;) { /* add point to clique */
	  /* The clique has k members (member[0..k-1]).
	     adjlist[0..k-1] point into the sorted adjacency lists.
	     Try to join a point >= nextPt into the clique. */
	  /* executed 109 times (vc9) */
	  for (;;) {
	    /* executed 1300 times (vc9) */
	    while (nextPt < np->vc_numPt && ! vc_active(np, nextPt)) nextPt ++;
	    if (nextPt >= np->vc_numPt) {
	      j = 0;
	      break;		/* no additional point */
	    }
#ifdef OLDLB4
	    for (j=0; j<k; j++) { /* find nextPt in adjlist[j] */
	      while (*adjlist[j] != ADJ_LAST &&
		     vc_pt(*adjlist[j]) < nextPt) adjlist[j] ++;
	      if (vc_pt(*adjlist[j]) != nextPt) break;
	    } /* for j */
#else /* OLDLB4 */
	    {
	      register MINT **adjj;
	      MINT **adjk;

	      adjj = adjlist;
	      adjk = adjlist + k;
	      for (; adjj < adjk; adjj++) { /* find nextPt in adjlist[j] */
		while (**adjj != ADJ_LAST &&
		       vc_pt(**adjj) < nextPt) (*adjj) ++;
		if (vc_pt(**adjj) != nextPt) break;
	      } /* for j */
	      j = adjj - adjlist;
	    }
#endif /* OLDLB4 */
	    if (j==k || *adjlist[j] == ADJ_LAST) break;
	    /* j < k */
	    nextPt = vc_pt(*adjlist[j]);
	    while (! vc_active(np, nextPt)) {
	      adjlist[j] ++;
	      if (*adjlist[j] == ADJ_LAST) break;
	      nextPt = vc_pt(*adjlist[j]);
	    }
	    if (*adjlist[j] == ADJ_LAST) break;
	  } /* for */
	  /* nextPt is in the clique or j < k */
	  if (j < k) break;
	  /*printf("Add %d\n", nextPt);*/
	  member[k] = nextPt;
	  adjlist[k] = vc_adjPt(global, nextPt);
	  k ++;
	  nextPt ++;
	} /* for: add point to clique */
	/* the clique is maximal */
	#ifndef ALLTRIANGLES
	  break;
	#endif
	if (k >= 3) break;	/* found at least a triangle */
	if (k <= 1) break;	/* did not find a second point */
	/* There is no triangle containing member[0] and member[1].
	   Try another point for member[1].
	   */
	adjlist[0] = vc_adjPt(global, i);
	k = 1;
	nextPt = member[1] + 1;
      } /* Try some second point */
      
      if (k >= 3) { /* mark */
#ifdef TRACELB
	printf("Clique (%d members):", k);
	for (j = 0; j < k; j++)
	  printf("%4d", member[j]);
	printf("\n");
#endif
	for (j = k-1; j > 0; j--)
	  vc_Mark(global, np, member[j], &head);
	if (vc_active(np, member[0])) {
	  vc_Mark(global, np, member[0], &head);
	  delta ++;
	} /* if active(member[0]) */
	MarkHeadTail(global, np, &head, &tail);
      } /* if: clique has at least three members -> mark */
    } /* for i: find clique starting with i and mark */

  return LowerBoundRest(global, np, head, tail) - delta;
} /* vc_LowerBound4 */

#define LB5a vc_LowerBound4

MINT vc_LowerBound5(vc_GlobalP global, vc_NodeP np)
{
  /* Partition the original graph into two subgraphs (left and right).
     Use a constant lower bound for both subgraphs. This gives
     a better lower bound near the root of the tree, but is of little use
     for 99% of the nodes.
     */
  int lbFull, lbLeft, lbRight;
  vc_NodeP np2;
  
  if (np->vc_numEdge == 0)
    return np->vc_numMarked;	/* This is a solution node */
  
  lbFull = LB5a(global, np);
  
  np2 = (vc_NodeP) z_NewCopyNode((z_NodeP) np);
  Restrict(global, np2, 1);
  lbLeft = LB5a(global, np2);
  (void) vc_FreeNode(np2);
  np2 = (vc_NodeP) z_NewCopyNode((z_NodeP) np);
  Restrict(global, np2, 0);
  lbRight = LB5a(global, np2);
  (void) vc_FreeNode(np2);
  
  /*printf("%c%c",
    (lbLeft <= global->vc_lbLeft) ? 'l' : 'L',
    (lbRight <= global->vc_lbRight) ? 'r' : 'R');*/
  /* Lower case -> effort was successful */
  if (lbLeft < global->vc_lbLeft) lbLeft = global->vc_lbLeft;
  if (lbRight < global->vc_lbRight) lbRight = global->vc_lbRight;
  /*printf("%c\n", (lbFull < lbLeft + lbRight) ? '!' : '.');*/
  /* ! -> effort was successful */
  /* . -> LowerBound4 is good enough */
  if (lbFull < lbLeft + lbRight) lbFull = lbLeft + lbRight;
  return lbFull;
} /* vc_LowerBound5 */

#define LB6a LB5a

static int lb6depth;

MINT vc_LowerBound6(vc_GlobalP global, vc_NodeP np)
{
  /* Partition the graph into two subgraphs (left and right).
     It is not sensible to partition the left subgraph into left and right.
     Must find a better criterion than act > 70.
     */
  int lbFull, lbLeft, lbRight, qsize;
  vc_NodeP np2, sol;
  
  if (np->vc_numEdge == 0)
    return np->vc_numMarked;	/* This is a solution node */

  if (lb6depth == 0 && np->vc_numAct > 70) {
    lbFull = LB6a(global, np);
    qsize = 10000000 / np->len;
    
    np2 = (vc_NodeP) z_NewCopyNode((z_NodeP) np);
    Restrict(global, np2, 1);	/* left */
    if (np2->vc_numAct == np->vc_numAct ||
	np2->vc_numAct == 0) {
      (void) vc_FreeNode(np2);
      return lbFull;
    }
    lb6depth ++;
    sol = (vc_NodeP) z_BranchBoundMixedFirst(global->vc_procsetnum,
					     (z_bb_GlobalDataP) global,
					     (z_NodeP)np2, NULL, qsize, 0);
    lbLeft = sol->vc_lb;
    (void) vc_FreeNode(np2);
    (void) vc_FreeNode(sol);
    
    np2 = (vc_NodeP) z_NewCopyNode((z_NodeP) np);
    Restrict(global, np2, 0);	/* right */
    sol = (vc_NodeP) z_BranchBoundMixedFirst(global->vc_procsetnum,
					     (z_bb_GlobalDataP) global,
					     (z_NodeP)np2, NULL, qsize, 0);
    lbRight = sol->vc_lb;
    (void) vc_FreeNode(np2);
    (void) vc_FreeNode(sol);
    
    /*printf("%c", (lbFull < lbLeft + lbRight) ? '!' : '.');
    fflush(stdout);*/
    /* ! -> effort was successful */
    /* . -> LowerBound4 is good enough */
    if (lbFull < lbLeft + lbRight) lbFull = lbLeft + lbRight;
    lb6depth --;
    return lbFull;
  } else
    return LB6a(global, np);
} /* vc_LowerBound6 */

#ifdef HISTO
#define lbMax 1000
static long lbHisto[lbMax];
static long actHisto[lbMax];
#endif /* HISTO */

MINT vc_LowerBound(vc_GlobalP global, vc_NodeP np)
{
  MINT lb;
  
  switch (global->vc_lbProc) {
  case 1:
    lb = vc_LowerBound1(global, np);
    break;
  case 2:
    lb = vc_LowerBound2(global, np);
    break;
  case 3:
    lb = vc_LowerBound3(global, np);
    break;
  case 4:
    lb = vc_LowerBound4(global, np);
    break;
  case 5:
    lb = vc_LowerBound5(global, np);
    break;
  case 6:
    lb = vc_LowerBound6(global, np);
    break;
#ifndef NOT_MAXIMAL_MATCHING
  case 7:
    lb = vc_LowerBound7(global, np);
    break;
  case 8:
    lb = vc_LowerBound8(global, np);
    break;
#endif
  default:
    lb = 0;			/* to avoid uninitialized variable warning */
    assert(0);
  }
#ifdef HISTO
  actHisto[np->vc_numAct] ++;
  lbHisto[lb] ++;
#endif /* HISTO */
  return lb;
}

void PrintLbHisto(char * name, int solution)
{
#ifdef HISTO
  int i;
  FILE *histofile;

  histofile = fopen("histo", "a");
  fprintf(histofile, "  lb   histo ------- %s\n", name);
  for (i=0;i<lbMax;i++)
    if (lbHisto[i])
      fprintf(histofile, "%4d%8ld%s\n", i, lbHisto[i],
	      (i == solution) ? " <- solution" : "");
  fprintf(histofile, " act   histo ------- %s\n", name);
  for (i=0;i<lbMax;i++)
    if (actHisto[i])
      fprintf(histofile, "%4d%8ld\n", i, actHisto[i]);
#endif /* HISTO */
}
